import hashlib
import six
import struct
import threading
import time
import transaction
import warnings
from copy import copy
from ZODB.DB import Pickler, BytesIO, _protocol, z64,\
        ConnectionPool, KeyedConnectionPool, IMVCCStorage
from ZODB.DB import DB as BaseDB
from ZEO.Exceptions import StorageError
from ZODB.POSException import POSKeyError
from ZODB.Connection import Connection as BaseConnection
from ZODB.Connection import RootConvenience
import base
from zerodb.storage import ServerStorage
from zerodb.transform.encrypt_common import get_encryption_signature, _gsm, IEncrypter


def create_root(storage, oid=z64, check_new=True):
    """
    Creates public or private root in storage.
    Root has the type PersistentMapping.

    :param storage: ZODB storage to create the root in
    :param str oid: Object id to give to the root (z64 is global root)
    :param bool check_new: If True, do nothing if the root exists
    """

    if check_new:
        try:
            storage.load(oid, '')
            return
        except KeyError:
            pass
    # Create the database's root in the storage if it doesn't exist
    from persistent.mapping import PersistentMapping
    root = PersistentMapping()
    # Manually create a pickle for the root to put in the storage.
    # The pickle must be in the special ZODB format.
    file = BytesIO()
    p = Pickler(file, _protocol)
    p.dump((root.__class__, None))
    p.dump(root.__getstate__())
    t = transaction.Transaction()
    t.description = 'initial database creation'
    storage.tpc_begin(t)
    storage.store(oid, None, file.getvalue(), '', t)
    storage.tpc_vote(t)
    storage.tpc_finish(t)


class StorageClass(ServerStorage):
    def set_database(self, database):
        """
        :param zerodb.permissions.base.PermissionsDatabase database: Database
        """
        assert isinstance(database, base.PermissionsDatabase)
        self.database = database
        self.noncekey = database.noncekey

    def _get_time(self):
        """Return a string representing the current time."""
        t = int(time.time())
        return struct.pack("i", t)

    def _get_nonce(self):
        s = ":".join([
            str(self.connection.addr),
            self._get_time(),
            self.noncekey])
        return hashlib.sha256(s).digest()

    def setup_delegation(self):
        # We use insert a hook to create a no-write root here
        super(StorageClass, self).setup_delegation()
        create_root(self.storage)

    def _check_permissions(self, data, oid=None):
        if not data.endswith(self.user_id):
            raise StorageError("Attempt to access encrypted data of others at <%s> by <%s>" % (oid, self.user_id.encode("hex")))

    def _check_admin(self):
        uid = struct.unpack(self.database.uid_pack, self.user_id)[0]
        user = self.database.db_root["users"][uid]
        assert user.administrator

    def loadEx(self, oid):
        data, tid = ServerStorage.loadEx(self, oid)
        self._check_permissions(data, oid)
        return data[:-len(self.user_id)], tid

    def load(self, oid, version=''):
        data, tid = ServerStorage.load(self, oid, version)
        self._check_permissions(data, oid)
        return data[:-len(self.user_id)], tid

    def loadBefore(self, oid, tid):
        r = ServerStorage.loadBefore(self, oid, tid)
        if r is not None:
            data, serial, after = r
            self._check_permissions(data, oid)
            return data[:-len(self.user_id)], serial, after
        else:
            return r

    def loadSerial(self, oid, serial):
        data = ServerStorage.loadSerial(self, oid, serial)
        self._check_permissions(data, oid)
        return data[:-len(self.user_id)]

    def storea(self, oid, serial, data, id):
        try:
            old_data, old_tid = ServerStorage.loadEx(self, oid)
            self._check_permissions(old_data, oid)
        except POSKeyError:
            pass  # We store a new one
        data += self.user_id
        return ServerStorage.storea(self, oid, serial, data, id)

    def get_root_id(self):
        """
        Gets 8-byte private root ID. Creates it if it doesn't exist.
        :return: <root_id>, <is root new>?
        :rtype: str, bool
        """
        uid = struct.unpack(self.database.uid_pack, self.user_id)[0]
        user = self.database.db_root["users"][uid]
        if user.root:
            try:
                self.storage.load(user.root, '')
                return user.root, False
            except KeyError:
                pass

        oid = self.storage.new_oid()
        with transaction.manager:
            user.root = oid
        return oid, True

    def add_user(self, username, pubkey, administrator=False):
        """
        Adminstrator can add a user
        """
        self._check_admin()
        self.database.add_user(username, pubkey, administrator=administrator)

    def del_user(self, username):
        """
        Adminstrator can remove a user
        """
        self._check_admin()
        self.database.del_user(username)

    def change_key(self, username, pubkey):
        """
        Administrator can change user's key (not reencrypting the data though!)
        """
        self._check_admin()
        self.database.change_key(username, pubkey)

    extensions = [get_root_id, add_user, del_user, change_key]

    # TODO
    # We certainly need to implement more methods for storage in here
    # (or check if they re-use loadEx):
    # deleteObject, storea, restorea, storeBlobEnd, storeBlobShared,
    # sendBlob, loadBulk


class Connection(BaseConnection):
    @property
    def root(self):
        return RootConvenience(self.get(self._db._root_oid))


class DB(BaseDB):
    klass = Connection

    # The __init__ method is largely replication of original ZODB.DB.DB.__init__
    # with the exception of root creation
    def __init__(self, storage,
                 pool_size=7,
                 pool_timeout=(1 << 31),
                 cache_size=400,
                 cache_size_bytes=0,
                 historical_pool_size=3,
                 historical_cache_size=1000,
                 historical_cache_size_bytes=0,
                 historical_timeout=300,
                 database_name='unnamed',
                 databases=None,
                 xrefs=True,
                 large_record_size=(1 << 24),
                 **storage_args):
        """Create an object database.

        :Parameters:
          - `storage`: the storage used by the database, e.g. FileStorage
          - `pool_size`: expected maximum number of open connections
          - `cache_size`: target size of Connection object cache
          - `cache_size_bytes`: target size measured in total estimated size
               of objects in the Connection object cache.
               "0" means unlimited.
          - `historical_pool_size`: expected maximum number of total
            historical connections
          - `historical_cache_size`: target size of Connection object cache for
            historical (`at` or `before`) connections
          - `historical_cache_size_bytes` -- similar to `cache_size_bytes` for
            the historical connection.
          - `historical_timeout`: minimum number of seconds that
            an unused historical connection will be kept, or None.
          - `xrefs` - Boolian flag indicating whether implicit cross-database
            references are allowed
        """
        storage_args = copy(storage_args)

        if isinstance(storage, six.string_types):
            import ZODB.FileStorage
            storage = ZODB.FileStorage.FileStorage(storage, **storage_args)
        elif storage is None:
            import ZODB.MappingStorage
            storage = ZODB.MappingStorage.MappingStorage(**storage_args)
        import ZODB  # Where did it go??

        # Allocate lock.
        x = threading.RLock()
        self._a = x.acquire
        self._r = x.release

        # pools and cache sizes
        self.pool = ConnectionPool(pool_size, pool_timeout)
        self.historical_pool = KeyedConnectionPool(historical_pool_size,
                                                   historical_timeout)
        self._cache_size = cache_size
        self._cache_size_bytes = cache_size_bytes
        self._historical_cache_size = historical_cache_size
        self._historical_cache_size_bytes = historical_cache_size_bytes

        # Setup storage
        self.storage = storage
        self.references = ZODB.serialize.referencesf
        try:
            storage.registerDB(self)
        except TypeError:
            storage.registerDB(self, None)  # Backward compat

        if (not hasattr(storage, 'tpc_vote')) and not storage.isReadOnly():
            warnings.warn(
                "Storage doesn't have a tpc_vote and this violates "
                "the storage API. Violently monkeypatching in a do-nothing "
                "tpc_vote.",
                DeprecationWarning, 2)
            storage.tpc_vote = lambda *args: None

        if IMVCCStorage.providedBy(storage):
            temp_storage = storage.new_instance()
        else:
            temp_storage = storage
        try:
            self._init_root(temp_storage, **storage_args)
        finally:
            if IMVCCStorage.providedBy(temp_storage):
                temp_storage.release()

        # Multi-database setup.
        if databases is None:
            databases = {}
        self.databases = databases
        self.database_name = database_name
        if database_name in databases:
            raise ValueError("database_name %r already in databases" %
                             database_name)
        databases[database_name] = self
        self.xrefs = xrefs

        self.large_record_size = large_record_size

    def _init_root(self, storage, **kw):
        oid, new = storage.get_root_id()
        storage._root_oid = oid
        if new:
            create_root(storage, oid=oid, check_new=False)
        elif hasattr(storage, "base"):
            # If a different encryption was used for this DB,
            # use that as default
            edata, _ = storage.base.load(oid)
            sig = get_encryption_signature(edata)
            utility = _gsm.getUtility(IEncrypter, name=sig)
            if _gsm.getUtility(IEncrypter) is not utility:
                _gsm.registerUtility(utility)
        self._root_oid = oid



"""
Combine multiple server storages in one class
"""
from types import MethodType
from ZEO.StorageServer import ZEOStorage


class ServerStorageMeta(type):
    def __init__(cls, name, bases, dct):
        dct = dict(dct)
        extensions = set(cls.extensions)
        for base in reversed(bases):
            if hasattr(base, "extensions"):
                extensions.update(base.extensions)
            else:
                for attr in dir(base):
                    meth = getattr(base, attr)
                    if not attr.startswith("_") and isinstance(meth, MethodType) and not hasattr(ZEOStorage, attr):
                        extensions.add(meth)
        cls.extensions = list(extensions)
        super(ServerStorageMeta, cls).__init__(name, bases, dct)


class MultiStorage(ZEOStorage, object):
    __metaclass__ = ServerStorageMeta



from zerodb.models import Model
from zerodb import intid


class T(Model):
    pass


def test_initid():
    ts = [T(), T(), T()]

    idstore = intid.IdStore()

    assert len(idstore) == 0

    for t in ts:
        idstore.add(t)

    assert len(idstore) == 3

    for t in ts:
        assert type(t._p_uid) in (int, long)

    assert len(set([t._p_uid for t in ts])) == 3
    assert idstore[ts[0]._p_uid] is ts[0]

    idstore.remove(ts[0])
    idstore.remove(ts[1]._p_uid)

    assert len(idstore) == 1



import pytest
from zerodb.crypto import rand
from zerodb.transform.encrypt_aes import AES256Encrypter as AES
from zerodb.crypto.exceptions import WrongKeyError

TEST_TEXT = "hello world"


def test_aes_rand():
    key = rand(32)

    cipher1 = AES(key=key)
    ciphertext = cipher1.encrypt(TEST_TEXT)

    cipher2 = AES(key=key)
    assert cipher2.decrypt(ciphertext) == TEST_TEXT


def test_aes_passphrase():
    passphrase = "the most secret passphrase ever"

    cipher1 = AES(passphrase=passphrase)
    ciphertext = cipher1.encrypt(TEST_TEXT)

    cipher2 = AES(passphrase=passphrase)
    assert cipher2.decrypt(ciphertext) == TEST_TEXT


def test_aes_exception():
    passphrase = "the most secret passphrase ever"

    cipher1 = AES(passphrase=passphrase)
    ciphertext = cipher1.encrypt(TEST_TEXT)

    cipher2 = AES(passphrase="wrong one")
    with pytest.raises(WrongKeyError):
        cipher2.decrypt(ciphertext)



from zope.component import getGlobalSiteManager
from zerodb.transform.interfaces import IEncrypter
from zerodb.transform.encrypt_common import get_encryption_signature


def test_root_encryption(db):
    _gsm = getGlobalSiteManager()
    utility = _gsm.getUtility(IEncrypter)
    edata, _ = db._connection._storage.base.load(db._connection._db._root_oid)
    assert get_encryption_signature(edata) == utility.name
    for obj in db._root.values():
        edata, _ = db._connection._storage.base.load(obj._p_oid)
        assert get_encryption_signature(edata) == ""



import pytest
import zerodb
from zerodb.transform.encrypt_aes import AES256Encrypter, AES256EncrypterV0
from conftest import do_zeo_server
from db import Page, TEST_PASSPHRASE, create_objects_and_close


class DBAESv0(zerodb.DB):
    encrypter = AES256EncrypterV0


@pytest.fixture(scope="module")
def zeo_server_aes_v0(request, pass_file, tempdir):
    # Use old pycryptodome encryption
    sock = do_zeo_server(request, pass_file, tempdir, dbclass=DBAESv0)
    create_objects_and_close(sock)
    return sock


@pytest.fixture(scope="module")
def db_aes_v0(request, zeo_server):
    # Use new zerodb.DB class here
    zdb = zerodb.DB(zeo_server, username="root", password=TEST_PASSPHRASE, debug=True)

    @request.addfinalizer
    def fin():
        zdb.disconnect()

    return zdb


def test_compat(db_aes_v0):
    # Test whether we can still read the DB
    assert len(db_aes_v0[Page]) > 0



from zerodb.collective.indexing.queue import getQueue
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent import ObjectModifiedEvent
from zope.event import subscribers

from zerodb.models import Model


def zerodb_autoreindex_dispatcher(event):
    if isinstance(event, ObjectModifiedEvent) and isinstance(event.object, Model):
        objectAutoReindex(event)


def objectAutoReindex(ev):
    indexer = getQueue()
    indexer.reindex(ev.object, ev.descriptions)   # put into queue, not really reindex.


def init():
    if zerodb_autoreindex_dispatcher not in subscribers:
        subscribers.append(zerodb_autoreindex_dispatcher)



from persistent import Persistent
from itertools import islice, izip, imap, count
from cachetools import LRUCache
from zerodb.storage import prefetch


class Sliceable(object):
    def __init__(self, f, cache_size=1000, length=None):
        """
        Makes a sliceable, cached list-like interface to an iterator
        :param callable f: Function which inits the iterator
        """
        self.f = f
        self.cache = LRUCache(cache_size)
        self.stop = 0
        self.length = length
        self.iterator = iter(f())

    def __iter__(self):
        for i in count():
            y = self.__getitem__(i)
            yield y
            # We get StopIteration error once we're done

    def dictify(self):
        for obj in self.__iter__():
            if hasattr(obj, "_p_activate"):
                obj._p_activate()
            yield {k: v for k, v in obj.__dict__.iteritems() if not k.startswith("_")}

    def __len__(self):
        if self.length is None:
            if hasattr(self.iterator, "__len__"):
                return len(self.iterator)
            else:
                return len(list(self.__iter__()))
        else:
            if callable(self.length):
                return self.length()
            else:
                return self.length

    def __getitem__(self, key):
        try:
            if isinstance(key, int) and (key >= 0):
                if key in self.cache:
                    return self.cache[key]
                elif key < self.stop:
                    self.stop = 0
                    self.iterator = iter(self.f())

                delta = key - self.stop
                result = islice(self.iterator, delta, delta + 1).next()
                self.cache[key] = result
                self.stop = key + 1
                return result

            elif isinstance(key, slice):
                if key.start is None and key.stop is None:
                    # Whole sequence is asked
                    return list(self.f())
                start = key.start or 0
                step = key.step or 1

                indexes = count(start, step)
                index_upd = start
                while (key.stop is None or index_upd < key.stop) and index_upd in self.cache:
                    index_upd += step

                if index_upd < self.stop and (key.stop is None or index_upd < key.stop):
                    self.iterator = iter(self.f())
                    result = list(islice(self.iterator, start, key.stop, step))
                    for i, value in izip(indexes, result):
                        self.cache[i] = value
                    self.stop = i + 1 if key.stop is None else key.stop
                    return result

                else:
                    result = [self.cache[i] for i in xrange(start, index_upd, step)]

                    if key.stop is None:
                        result_upd = list(islice(self.iterator, index_upd - self.stop, None, step))
                    elif index_upd < key.stop:
                        result_upd = list(islice(self.iterator, index_upd - self.stop, key.stop - self.stop, step))
                    else:
                        result_upd = []
                    for i, value in izip(indexes, result_upd):
                        self.cache[i] = value
                    self.stop = key.stop
                    return result + result_upd

            else:
                raise KeyError("Key must be non-negative integer or slice, not {}"
                               .format(key))

        except StopIteration:
            self.iterator = self.f()
            self.stop = 0
            raise

    def __repr__(self):
        """ Visually appealing output showing first 5 elements of the data """
        first_el = self[:6]
        is_long = len(first_el) > 5
        reprs = [i.__repr__() for i in first_el[:5]]
        if is_long:
            reprs.append("...")
        if len(reprs) <= 1:
            return "[" + "".join(reprs) + "]"
        else:
            l = len(reprs)
            out = []
            for i, s in enumerate(reprs):
                if i == 0:
                    s = "[" + s
                else:
                    s = " " + s
                if i == l - 1:
                    s = s + "]"
                else:
                    s = s + ","
                out.append(s)
            return "\n".join(out)

    def __unicode__(self):
        return self.__repr__()


class DBList(Sliceable):
    def __init__(self, query_f, db, **kw):
        """
        :param function query_f: Function which returns results of the query in format (size, uids)
        :param zerodb.DB db: Currend DB instance
        """
        self.db = db

        def get_object(uid):
            obj = db._objects[uid]
            obj._p_uid = uid
            return obj

        def f():
            self.length, it = query_f()
            return imap(get_object, it)

        super(DBList, self).__init__(f, **kw)


class ListPrefetch(Sliceable):
    prefetch_size = 20

    def __getitem__(self, key):
        previous_stop = self.stop
        result = super(ListPrefetch, self).__getitem__(key)
        if self.stop != previous_stop:
            # Cache-ahead
            try:
                if isinstance(key, (int, long)):
                    tail = super(ListPrefetch, self).__getitem__(slice(key + 1, key + self.prefetch_size + 1))
                elif isinstance(key, slice):
                    if key.stop:
                        tail = super(ListPrefetch, self).__getitem__(slice(key.stop + 1, key.stop + self.prefetch_size + 1))
                    else:
                        tail = []
            except StopIteration:
                # If we've finished right at this element, that's not an error
                tail = []

            # Fetching objects needed
            if isinstance(result, list):
                prefetch(result + tail)
            elif isinstance(result, Persistent):
                prefetch([result] + tail)
        return result


class DBListPrefetch(ListPrefetch, DBList):
    def __init__(self, query_f, db, **kw):
        DBList.__init__(self, query_f, db, **kw)



from repoze.catalog.catalog import ResultSetSize
from repoze.catalog.catalog import Catalog as _Catalog
from zerodb import trees


class Catalog(_Catalog):
    family = trees.family32

    def sort_result(self, result, sort_index=None, limit=None, sort_type=None,
                    reverse=False):

        if sort_index:
            result = set(result)
            numdocs = total = len(result)
            index = self[sort_index]
            result = index.sort(result, reverse=reverse, limit=limit,
                                sort_type=sort_type)
            if limit:
                numdocs = min(numdocs, limit)
            size = ResultSetSize(numdocs, total)
        else:
            if limit:
                size = ResultSetSize(limit, None)
            else:
                size = None

        return size, result



from zerodb.storage import multi


def test_multistorage():
    class X(object):
        def _x(self):
            pass

        def x(self):
            pass

        def loadEx(self, x):
            return x

    class Y(object):
        def y(self):
            pass

        def z(self):
            pass

        extensions = [y]

    class M(X, Y, multi.MultiStorage):
        pass

    assert len(M.extensions) == 2
    m = M(None)
    m.x()
    m.y()
    assert m.loadEx(42) == 42



import logging
import mock
import transaction
from itertools import islice
from db import Page, Salary, Department
from zerodb.catalog.query import Contains, InRange, Eq, Gt
# Also need to test optimize, Lt(e), Gt(e)

logging.basicConfig(level=logging.DEBUG)


def test_query(db):
    pre_request_count = db._storage._debug_download_count
    assert len(db[Page]) == 201
    assert len(db[Salary]) == 201
    test_pages = db[Page].query(Contains("text", "something"))
    pre_range_count = db._storage._debug_download_count
    assert pre_range_count - pre_request_count < 15  # We'll have performance testing separately this way
    test_salaries_1 = db[Salary].query(InRange("salary", 130000, 180000), sort_index="salary", limit=2)
    test_salaries_2 = db[Salary].query(InRange("salary", 130000, 130001), sort_index="salary", limit=2)
    post_range_count = db._storage._debug_download_count
    assert len(test_pages) == 10
    assert len(test_salaries_1) == 2
    assert len(test_salaries_2) == 0
    for s in test_salaries_1:
        assert s.salary >= 130000
        assert s.salary <= 180000
    # Check that we pre-downloaded all objects into cache
    assert db._storage._debug_download_count - post_range_count <= 5

    million_employee = db[Salary].query(salary=1000000)
    assert len(million_employee) == 1
    assert million_employee[0].name == "Hello"

    assert len(db[Salary].query(full_name="Hello World")) == 1
    assert len(db[Salary].query(full_name="Hell? Wor*")) == 1
    assert db[Salary].query(Contains("full_name", "Hello"))[0] == million_employee[0]

    assert db[Salary].query(future_salary=2000000)[0] == million_employee[0]

    assert len(db[Salary].query(department_name="Mobile")) > 0
    department = db[Department].query(name="Money")[0]
    assert len(db[Salary].query(department_id=department._p_uid)) > 0

    test_pages = db[Page].query(Contains("text", "something"))[:5]
    assert len(test_pages) == 5

    db[Salary].query(InRange("salary", 130000, 180000), name="John-2")
    db[Salary].query(InRange("salary", 130000, 180000) | Eq("name", "John-2"))

    # test Length object
    delattr(db[Salary]._objects, "length")
    assert len(db[Salary]) > 0
    assert db[Salary]._objects.length.value == len(db[Salary])


def test_add(db):
    with transaction.manager:
        pre_commit_count = db._storage._debug_download_count
        page = Page(title="hello", text="Quick brown lazy fox jumps over lorem  ipsum dolor sit amet")
        db.add(page)
        post_commit_count = db._storage._debug_download_count
    print "Number of requests:", post_commit_count - pre_commit_count
    assert post_commit_count - pre_commit_count < 22

    with transaction.manager:
        db.remove(page)


def test_reindex(db):
    with transaction.manager:
        page = Page(title="hello", text="Quick0 brown lazy fox jumps over lorem  ipsum dolor sit amet")
        docid = db.add(page)
    assert len(db[Page].query(Contains("text", "quick0"))) == 1

    # DbModel, by ID
    with transaction.manager:
        page.text = "Quick1 brown lazy fox jumps over well, you know"
        db[Page].reindex(docid)
    assert len(db[Page].query(Contains("text", "quick0"))) == 0
    assert len(db[Page].query(Contains("text", "quick1"))) == 1

    # DbModel, by obj
    with transaction.manager:
        page.text = "quick2 brown lazy fox jumps over well, you know"
        db[Page].reindex(page)
    assert len(db[Page].query(Contains("text", "quick1"))) == 0
    assert len(db[Page].query(Contains("text", "quick2"))) == 1

    # DB, by obj
    with transaction.manager:
        page.text = "quick3 brown lazy fox jumps over well, you know"
        db[Page].reindex(page)
    assert len(db[Page].query(Contains("text", "quick2"))) == 0
    assert len(db[Page].query(Contains("text", "quick3"))) == 1

    # DB, multiple objects
    with transaction.manager:
        page2 = Page(title="hello", text="Quick4 brown lazy fox jumps over lorem  ipsum dolor sit amet")
        db.add(page2)

    with transaction.manager:
        page.text = "quick5 brown lazy fox jumps over well, you know"
        page2.text = "quick5 brown lazy fox jumps over well, you know"
        db.reindex([page, page2])
    assert len(db[Page].query(Contains("text", "quick3") | Contains("text", "quick4"))) == 0
    assert len(db[Page].query(Contains("text", "quick5"))) == 2


def test_auto_reindex(db):
    with transaction.manager:
        page = Page(title="hello", text="autoreindex0, test whether to work")
        db.add(page)
    assert len(db[Page].query(Contains("text", "autoreindex0"))) == 1

    with transaction.manager:
        page.text = "autoreindex1, test whether to work"
    assert len(db[Page].query(Contains("text", "autoreindex0"))) == 0
    assert len(db[Page].query(Contains("text", "autoreindex1"))) == 1

    with transaction.manager:
        page2 = Page(title="hello", text="autoreindex2, test whether to work")
        db.add(page2)

    with transaction.manager:
        page.text = "autoreindex3, test whether to work"
        page2.text = "autoreindex3, test whether to work"
    assert len(db[Page].query(Contains("text", "autoreindex1") | Contains("text", "autoreindex2"))) == 0
    assert len(db[Page].query(Contains("text", "autoreindex3"))) == 2

    with mock.patch("zerodb.db.DbModel.reindex_one") as reindex_mock:
        with transaction.manager:
            page.text = "autoreindex3, test whether to work1"
            page.text = "autoreindex3, test whether to work2"
            page.text = "autoreindex3, test whether to work3"
        assert reindex_mock.call_count == 1

    db.enableAutoReindex(False)
    with transaction.manager:
        page.text = "autoreindex4, test whether to work"
    assert len(db[Page].query(Contains("text", "autoreindex3"))) == 2
    assert len(db[Page].query(Contains("text", "autoreindex4"))) == 0

    db.enableAutoReindex(True)
    with transaction.manager:    # should not throw ModleException
        page3 = Page(title="helloworld", text="autoreindex5, test whether to work")
        page3.title = "helloworld1"
    assert len(db[Page].query(Eq("title", "helloworld"))) == 0
    assert len(db[Page].query(Eq("title", "helloworld1"))) == 0

    with mock.patch("zerodb.db.DbModel.reindex_one") as reindex_mock:
        with transaction.manager:  # should not reindex
            page3 = Page(title="helloworld", text="autoreindex5, test whether to work")
            page3.title = "helloworld1"
            db.add(page3)
        assert reindex_mock.call_count == 0

    with transaction.manager:  # should  reindex
        page3 = Page(title="helloworld", text="autoreindex6, test whether to work")
        db.add(page3)
        page3.title = "helloworld1"
        page3.text = "autoreindex7, test whether to work"
    assert len(db[Page].query(Eq("title", "helloworld"))) == 0
    assert len(db[Page].query(Eq("title", "helloworld1"))) == 2
    assert len(db[Page].query(Contains("text", "autoreindex6"))) == 0
    assert len(db[Page].query(Contains("text", "autoreindex7"))) == 1

    with mock.patch("zerodb.db.DbModel.reindex_one") as reindex_mock:
        with transaction.manager:  # should  reindex
            page3 = Page(title="helloworld", text="autoreindex6, test whether to work")
            db.add(page3)
            page3.title = "helloworld1"
            page3.text = "autoreindex7, test whether to work"
        assert reindex_mock.call_count == 1


def test_fieldindex_typechange(db):
    with transaction.manager:
        for i in range(30):
            db.add(Salary(
                name="Tom",
                surname="Jackson-%s" %i,
                salary=10000 + 100 * i))
    test1 = db[Salary].query(InRange("salary", 9999, 14000))
    assert len(test1) == 30
    test2 = db[Salary].query(InRange("salary", 9999, 10901) & Eq("name", "Tom"))
    assert len(test2) == 10


def test_repr(db):
    data = db[Salary].query(Gt("salary", 100000))
    s = str(data)
    assert s.startswith("[")
    assert "..." in s
    assert s.endswith("]")
    lines = s.split("\n")
    assert sum([l.endswith(",") for l in lines]) == 5
    assert sum([l.startswith(" ") for l in lines]) == 5

    data = db[Salary].query(Gt("salary", 100000), name="Nonexistent")
    assert str(data) == "[]"


def test_pack(db):
    db.pack()
    assert len(db[Page]) > 0


def test_all_uid(db):
    # Test for https://gist.github.com/micxjo/a097698b33fc4669b0b4
    page = Page(title="Test page", text="Hello world")
    with transaction.manager:
        db.add(page)

    del page
    # Clear in-memory and on-disk caches
    db._storage._cache.clear()
    db._connection._cache.full_sweep()

    for item in db[Page].all():
        assert hasattr(item, "_p_uid")
        del item

    db._storage._cache.clear()
    db._connection._cache.full_sweep()

    for uid in db[Page].all_uids():
        obj = db[Page][uid]
        assert hasattr(obj, "_p_uid")
        del obj

    db._storage._cache.clear()
    db._connection._cache.full_sweep()

    uids = list(islice(db[Page].all_uids(), 10))
    objs = db[Page][uids]
    for obj in objs:
        assert hasattr(obj, "_p_uid")



from zc.zlibstorage import ZlibStorage
import logging
import zope.interface
from zerodb.transform import encrypt, decrypt, compress, decompress
from zerodb.util.debug import debug_loads


class TransformingStorage(ZlibStorage):
    """
    Storage which can transform (encrypt and/or compress) data.
    Also this storge is aware of our loadBulk method
    """

    def __init__(self, base, *args, **kw):
        """
        :param base: Storage to transform
        :param bool debug: Output debug log messages
        """
        self.base = base

        self.debug = kw.pop("debug", False)
        if self.debug:
            self._debug_download_size = 0
            self._debug_download_count = 0

        for name in self.copied_methods:
            v = getattr(base, name, None)
            if v is not None:
                setattr(self, name, v)

        zope.interface.directlyProvides(self, zope.interface.providedBy(base))

        base.registerDB(self)

        self._transform = lambda data: encrypt(compress(data), no_cipher_name=True)
        self._transform_named = lambda data: encrypt(compress(data), no_cipher_name=False)
        self._untransform = lambda data: decompress(decrypt(data))

        self._root_oid = None

    def load(self, oid, version=''):
        """
        Load object by oid

        :param str oid: Object ID
        :param version: Version to load (when we have version control)
        :return: Object and its serial number
        :rtype: tuple
        """
        if self.debug:
            if oid not in self._cache.current:
                in_cache = False
            else:
                in_cache = True

        data, serial = self.base.load(oid, version)
        out_data = self._untransform(data)

        if self.debug and not in_cache:
            logging.debug("id:%s, type:%s, transform: %s->%s" % (oid.encode("hex"), debug_loads(out_data), len(data), len(out_data)))
            self._debug_download_size += len(data)
            self._debug_download_count += 1

        return out_data, serial

    def loadBulk(self, oids, returns=True):
        """
        Load multiple objects at once

        :param list oids: Iterable of oids to load
        :param bool returns: When False, we don't return objects but store them
            in cache
        :return: List of (object, serial) tuples
        :rtype: list
        """
        if self.debug:
            logging.debug("Loading: " + ", ".join([oid.encode("hex") for oid in oids]))
            in_cache_before = {oid: oid in self._cache.current for oid in oids}
        base_result = self.base.loadBulk(oids)
        if self.debug:
            in_cache_after = {oid: oid in self._cache.current for oid in oids}
        if returns or self.debug:
            datas, serials = zip(*base_result)
            datas_out = map(self._untransform, datas)
            out = zip(datas_out, serials)
            if self.debug:
                if datas:
                    self._debug_download_count += 1
                for data, out_data, oid in zip(datas, datas_out, oids):
                    logline_prefix = ""
                    if not in_cache_before[oid]:
                        self._debug_download_size += len(data)
                        if not in_cache_after[oid]:
                            self._debug_download_count += 1
                        else:
                            logline_prefix = "(from bulk) "
                        logging.debug("%sid:%s, type:%s, transform: %s->%s" %
                                (logline_prefix, oid.encode("hex"), debug_loads(out_data), len(data), len(out_data)))
            if returns:
                return out

    def store(self, oid, serial, data, version, transaction):
        if oid == self._root_oid:
            _transform = self._transform_named
        else:
            _transform = self._transform
        return self.base.store(oid, serial, _transform(data), version,
                               transaction)



import pytest
import logging
import requests
import socket
import time
from json import loads, dumps
from multiprocessing import Process
from os import path
from db import TEST_PASSPHRASE
from zerodb import api


logging.basicConfig(level=logging.DEBUG)


def api_run(**kw):
    api.run(use_reloader=False, **kw)


@pytest.fixture(scope="module")
def api_server(request, db):
    # Get available TCP port
    sock = socket.socket()
    sock.bind(("localhost", 0))
    _, port = sock.getsockname()
    sock.close()

    server = Process(target=api_run, kwargs={
        "host": "localhost",
        "port": port,
        "data_models": path.join(path.dirname(__file__), "db.py"),
        "zeo_socket": db._storage._addr})

    @request.addfinalizer
    def fin():
        server.terminate()
        server.join()

    server.start()
    time.sleep(0.2)

    return "http://localhost:%s" % port


def api_connect(api_server, session):
    return session.get(api_server + "/_connect", params={
        "username": "root",
        "passphrase": TEST_PASSPHRASE})


def api_disconnect(api_server, session):
    return session.get(api_server + "/_disconnect")


def test_connect(api_server):
    session = requests.Session()

    # Connect
    resp = api_connect(api_server, session)
    assert loads(resp.text)["ok"] == 1

    # Disconnect
    resp = api_disconnect(api_server, session)
    assert loads(resp.text)["ok"] == 1


def test_insert_get(api_server):
    session = requests.Session()
    docs = [{"title": "test inserting one", "text": "Here we go, test insert"},
            {"title": "test inserting two", "text": "What's going on here?"}]

    api_connect(api_server, session)

    resp = session.post(api_server + "/Page/_insert",
            data={"docs": dumps(docs)})
    resp = loads(resp.text)

    assert resp["status"]["ok"] == 1
    oids = [o["$oid"] for o in resp["oids"]]
    assert all(oids)

    resp = session.post(api_server + "/Page/_get", data={"_id": dumps(oids)})
    resp = loads(resp.text)
    assert resp == docs

    api_disconnect(api_server, session)


def test_find(api_server):
    session = requests.Session()
    api_connect(api_server, session)

    resp = session.post(api_server + "/Page/_find", data={
        "criteria": dumps({"text": {"$text": "something"}})
        })
    resp = loads(resp.text)
    assert len(resp) == 10

    resp = session.post(api_server + "/Salary/_find", data={
        "criteria": dumps({"salary": {"$range": [130000, 180000]}}),
        "limit": 2,
        "sort": "salary"
        })
    resp = loads(resp.text)
    assert len(resp) == 2

    resp = session.post(api_server + "/Salary/_find", data={
        "criteria": dumps({"salary": {"$range": [130000, 130001]}}),
        "limit": 2,
        "sort": {"salary": -1}
        })
    resp = loads(resp.text)
    assert len(resp) == 0

    api_disconnect(api_server, session)


def test_remove_by_id(api_server):
    docs = [{"title": "test removing one", "text": "Something to remove 1"},
            {"title": "test removing two", "text": "Something to remove 2"},
            {"title": "test removing three", "text": "Something to remove 3"}]
    session = requests.Session()
    api_connect(api_server, session)

    resp = session.post(api_server + "/Page/_insert",
            data={"docs": dumps(docs)})
    resp = loads(resp.text)
    oids = [o["$oid"] for o in resp["oids"]]

    resp = session.post(api_server + "/Page/_remove",
            data={"_id": dumps(oids)})
    resp = loads(resp.text)
    assert resp["ok"] == 1
    assert resp["count"] == 3

    resp = session.post(api_server + "/Page/_find", data={
        "criteria": dumps({"text": {"$text": "something remove"}})
        })
    resp = loads(resp.text)
    assert len(resp) == 0

    api_disconnect(api_server, session)


def test_remove_by_criteria(api_server):
    docs = [{"title": "test removing one", "text": "Something to remove 1 placeholder"},
            {"title": "test removing two", "text": "Something to remove 2 placeholder"},
            {"title": "test removing three", "text": "Something to delete 3 placeholder"}]
    session = requests.Session()
    api_connect(api_server, session)

    resp = session.post(api_server + "/Page/_insert",
            data={"docs": dumps(docs)})

    resp = session.post(api_server + "/Page/_remove", data={
            "criteria": dumps({"text": {"$text": "something remove"}})})
    resp = loads(resp.text)
    assert resp["ok"] == 1
    assert resp["count"] == 2

    resp = session.post(api_server + "/Page/_find", data={
        "criteria": dumps({"text": {"$text": "something placeholder"}})
        })
    resp = loads(resp.text)
    assert len(resp) == 1

    api_disconnect(api_server, session)



from zerodb.transform import compress, decompress
from zerodb.transform.compress_zlib import zlib_compressor


def test_utilities():
    zlib_compressor.register(default=True)
    test_string = "This is a test " * 1000

    compressed_default = compress(test_string)
    compressed_zlib = zlib_compressor.compress(test_string)

    assert compressed_default == compressed_zlib
    assert compressed_default != test_string
    assert compressed_zlib != test_string

    assert decompress(compressed_default) == test_string
    assert decompress(compressed_zlib) == test_string



from ZEO.StorageServer import StorageServer as BaseStorageServer
from ZEO.runzeo import ZEOServer as BaseZEOServer
from ZEO.runzeo import ZEOOptions
import ZEO.zrpc.error
from itertools import chain
from persistent import Persistent

import batch
import premade
import transforming
import logging

# TODO when it comes to the point we need to,
# we'll have to configure which classes to use
# with Zope interfaces

ServerStorage = premade.DefaultServerStorage


class StorageServer(BaseStorageServer):
    ZEOStorageClass = ServerStorage

    def invalidate(self, conn, storage_id, tid, invalidated=(), info=None):
        """ Internal: broadcast info and invalidations to clients. """

        # Same as in the base class
        if invalidated:
            invq = self.invq[storage_id]
            if len(invq) >= self.invq_bound:
                invq.pop()
            invq.insert(0, (tid, invalidated))

        for p in self.connections[storage_id]:
            if getattr(p, 'user_id', None) == getattr(conn, 'user_id', None):
                try:
                    if invalidated and p is not conn:
                        p.client.invalidateTransaction(tid, invalidated)
                    elif info is not None:
                        p.client.info(info)
                except ZEO.zrpc.error.DisconnectedError:
                    pass


class ZEOServer(BaseZEOServer):
    def create_server(self):
        storages = self.storages
        options = self.options
        self.server = StorageServer(
            options.address,
            storages,
            read_only=options.read_only,
            invalidation_queue_size=options.invalidation_queue_size,
            invalidation_age=options.invalidation_age,
            transaction_timeout=options.transaction_timeout,
            monitor_address=options.monitor_address,
            auth_protocol=options.auth_protocol,
            auth_database=options.auth_database,
            auth_realm=options.auth_realm)

    @classmethod
    def run(cls, args=None):
        options = ZEOOptions()
        options.realize(args=args)

        for o_storage in options.storages:
            if o_storage.config.pack_gc:
                logging.warn("Packing with GC and end-to-end encryption removes all the data")
                logging.warn("Turining GC off!")
                o_storage.config.pack_gc = False

        s = cls(options)
        s.main()


def client_storage(sock, *args, **kw):
    """
    Storage client

    :param sock: UNIX or TCP socket
    :param cipher: Encryptor to use (see zerodb.crypto)
    :param bool debug: Output debug messages to the log
    :param transforming_storage: Wrapper-storage to use (TransformingStorage)
    :returns: Storage
    :rtype: TransformingStorage
    """
    if type(sock) is unicode:
        sock = str(sock)
    TransformingStorage = kw.pop('transforming_storage', transforming.TransformingStorage)
    debug = kw.pop("debug", False)
    return TransformingStorage(batch.BatchClientStorage(sock, *args, **kw), debug=debug)


def prefetch(objs):
    """
    Bulk-fetch ZODB objects
    """
    objs = filter(lambda x: hasattr(x, "_p_oid"), objs)
    if objs:
        oids = [y._p_oid for y in objs if y._p_oid is not None]
        if objs[0]._p_jar:
            objs[0]._p_jar._db._storage.loadBulk(oids, returns=False)


def prefetch_trees(trees, depth=10, bucket_types=(), shallow=True):
    """
    Bulk-fetch specified trees or buckets in logarithmic number of steps
    """
    trees = [t for t in trees if isinstance(t, Persistent) and hasattr(t, "_p_oid")]

    if len(trees) == 0 or depth == 0:
        # If someone inserts an infinite loop, we shouldn't go for that
        return

    if shallow:  # For performance once we have everything in cache
        cache_load = trees[0]._p_jar._db._storage._cache.load
        if not [True for o in trees if cache_load(o._p_oid) is None]:
            return

    prefetch(trees)

    if not bucket_types:
        bucket_types = tuple(set(chain(*[(type(t), t._bucket_type) for t in trees if hasattr(t, "_bucket_type")])))

    children = []
    for tree in trees:
        state = tree.__getstate__()
        if state:
            children += [o for o in state[0] if isinstance(o, bucket_types)]

    prefetch_trees(children, depth=(depth - 1), bucket_types=bucket_types)


def btree_state_search(state, key):
    """
    Search in raw state of BTree.
    state = ((branch, key, branch, key, ... , key, branch), 1st_bucket) or None
    """
    if not state:
        return -1, None

    L = (len(state[0]) + 1) / 2

    get_key = lambda i: state[0][i * 2 - 1] if i > 0 else None

    # Copy of python BTree logic
    lo = 0
    hi = L
    i = hi // 2
    while i > lo:
        cmp_ = cmp(get_key(i), key)
        if cmp_ < 0:
            lo = i
        elif cmp_ > 0:
            hi = i
        else:
            break
        i = (lo + hi) // 2
    return i, state[0][i * 2]


def parallel_traversal(trees, keys):
    """
    Traverse trees in parallel to fill up cache
    """
    if not isinstance(trees, (list, tuple)):
        to_fetch = [trees]
        trees = [trees] * len(keys)
    else:
        to_fetch = list(set(t for t in trees if isinstance(t, Persistent) and hasattr(t, "_p_oid")))

    prefetch(to_fetch)

    nxt_trees = []
    nxt_keys = []
    for key, tree in zip(keys, trees):
        if isinstance(tree, Persistent) and hasattr(tree, "_bucket_type"):
            _, nxt = btree_state_search(tree.__getstate__(), key)
            if isinstance(nxt, (type(tree), tree._bucket_type)):
                nxt_keys.append(key)
                nxt_trees.append(nxt)

    if nxt_keys:
        parallel_traversal(nxt_trees, nxt_keys)



"""
BTrees with modified bucket sizes
"""

import BTrees


class TreeFamily32(BTrees._Family32):
    import IFBTree as IF
    import IIBTree as II
    import IOBTree as IO
    import OIBTree as OI
    import OOBTree as OO


family32 = TreeFamily32()
family32.IF.family = family32
family32.II.family = family32
family32.IO.family = family32
family32.OI.family = family32
family32.OO.family = family32



import itertools
import pytest

from zerodb.util.iter import Sliceable
from zerodb.catalog.query import Gt
from db import Salary


def test_sliceable():
    it = Sliceable(lambda: itertools.imap(str, itertools.count()))
    assert it[500] == "500"
    assert it[505] == "505"
    it1 = it.iterator
    it.cache.clear()
    assert it[0] == "0"
    assert it[2] == "2"
    assert it1 is not it.iterator

    assert it[10:15:3] == map(str, range(10, 15, 3))
    assert it[101:200:5] == map(str, range(101, 200, 5))
    assert it[20:25:3] == map(str, range(20, 25, 3))

    assert it[5:100] == map(str, range(5, 100))
    it1 = it.iterator
    assert it[10:20] == map(str, range(10, 20))
    assert it.iterator is it1

    with pytest.raises(KeyError):
        it[-1]
    with pytest.raises(KeyError):
        it["raise"]

    it = Sliceable(lambda: xrange(10))

    assert len([i for i in it]) == 10
    assert len(it) == 10

    it = Sliceable(lambda: itertools.imap(str, xrange(100)))
    assert len(it) == 100
    it = Sliceable(lambda: itertools.imap(str, xrange(100)))
    assert it[10:] == map(str, range(10, 100))


def test_dictify(db):
    test_salaries = db[Salary].query(Gt("salary", 100000)).dictify()
    obj = test_salaries.next()
    assert set(obj.keys()) == set(["name", "surname", "salary", "department"])



import pytest
from ZEO.Exceptions import StorageError
from ZODB.DB import z64
from zerodb.permissions import subdb
from zerodb.crypto import ecc
from zerodb.storage import client_storage
from conftest import TEST_PUBKEY, TEST_PASSPHRASE

import logging
logging.basicConfig(level=logging.DEBUG)


def test_db_rootuser(pass_db):
    user = pass_db["root"]
    assert user.administrator
    assert user.pubkey == TEST_PUBKEY


def test_db_users(pass_db):
    pk1 = ecc.private("pass1").get_pubkey()
    pk2 = ecc.private("pass2").get_pubkey()
    pk3 = ecc.private("pass3").get_pubkey()
    pass_db.add_user("user1", pk1, administrator=True)
    pass_db.add_user("user2", pk2)
    pass_db.add_user("user3", pk3)
    pass_db.del_user("user3")
    pass_db.change_key("user2", pk3)

    assert pass_db["user1"].administrator
    assert not pass_db["user2"].administrator
    assert pass_db["user1"].pubkey == pk1
    assert pass_db["user2"].pubkey == pk3

    with pytest.raises(LookupError):
        pass_db["user3"]


def test_ecc_auth(zeo_server):
    # Presumably, ecc_server already registered auth protocol
    storage = client_storage(zeo_server,
            username="root", password=TEST_PASSPHRASE, realm="ZERO")

    with pytest.raises(StorageError):  # Cannot access common root
        storage.load(z64)

    db = subdb.DB(storage)
    conn = db.open()
    conn.root()

    assert db._root_oid != z64
    assert type(db._root_oid) == str

    conn.close()


def test_user_management(zeo_server):
    storage = client_storage(zeo_server,
            username="root", password=TEST_PASSPHRASE, realm="ZERO")

    pk0 = ecc.private("passY").get_pubkey()
    pk = ecc.private("passX").get_pubkey()
    storage.add_user("userX", pk0)
    storage.change_key("userX", pk)

    storage = client_storage(zeo_server,
            username="userX", password="passX", realm="ZERO")
    with pytest.raises(AssertionError):
        storage.add_user("shouldfail", pk)



# Taken from:
# https://gist.github.com/ajdavis/4644641/

from __future__ import print_function

import threading
import weakref

from functools import partial


class ThreadWatcher(object):
    class Vigil(object):
        pass

    def __init__(self):
        self._refs = {}
        self._local = threading.local()

    def _on_death(self, vigil_id, callback, args, ref):
        self._refs.pop(vigil_id)
        callback(*args)

    def watch(self, callback, *args):
        if not self.is_watching():
            self._local.vigil = v = ThreadWatcher.Vigil()
            on_death = partial(
                self._on_death, id(v), callback, args)

            ref = weakref.ref(v, on_death)
            self._refs[id(v)] = ref

    def is_watching(self):
        "Is the current thread being watched?"
        try:
            v = self._local.vigil
            return id(v) in self._refs
        except AttributeError:
            return False

    def unwatch(self):
        try:
            v = self._local.vigil
            del self._local.vigil
            self._refs.pop(id(v))
        except AttributeError:
            pass



#!/usr/bin/env python

# Warning!
# At the moment, this API is safe to use only on same machine
# (no https etc)

import imp
import jsonpickle
import transaction
import zerodb
from flask import Flask, jsonify, session, request
from flask.ext.cors import CORS
from zerodb.catalog.query import optimize
from zerodb.catalog import query_json as qj

try:
    import simplejson as json
except:
    import json

PORT = 2015
HOST = "127.0.0.1"
DEBUG = True
DEV_SECRET_KEY = "development key"

app = Flask(__name__)
cors = CORS(app, supports_credentials=True)
dbs = {}
models = None
zeosocket = None


@app.route("/_connect", methods=["GET", "POST"])
def connnect():
    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    username = req.get("username")
    passphrase = req.get("passphrase")
    if zeosocket:
        socket = zeosocket
    else:
        host = req.get("host")
        port = req.get("port")
        if host and port:
            socket = (host, port)
        elif host:
            socket = host
        else:
            socket = None

    if not (username and passphrase and socket):
        return jsonify(ok=0, message="Incomplete login information")

    try:
        db = zerodb.DB(socket, username=username, password=passphrase)
    except Exception, e:
        return jsonify(ok=0, message=str(e), error_type=e.__class__.__name__)

    session["username"] = username
    dbs[username] = db

    return jsonify(ok=1)


@app.route("/<table_name>/_find", methods=["GET", "POST"])
def find(table_name):
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    if req.get("criteria") is None:
        criteria = json.loads(request.data).get("criteria")
    else:
        criteria = json.loads(req.get("criteria"))

    if isinstance(criteria, dict) and (len(criteria) == 1) and "_id" in criteria:
        ids = [c["$oid"] for c in criteria["_id"]]
    else:
        ids = None
        criteria = optimize(qj.compile(criteria))

    skip = req.get("skip")
    if skip:
        skip = int(skip)

    limit = req.get("limit")
    if limit:
        limit = int(limit)

    sort = req.get("sort")
    if sort:
        try:
            sort = json.loads(sort)
        except ValueError:
            if sort.startswith("-"):
                sort_index = sort[1:].strip()
                reverse = True
            else:
                sort_index = sort
                reverse = None
        if isinstance(sort, dict):
            assert len(sort) == 1  # Only one field at the moment
            sort_index, direction = sort.popitem()
            reverse = (direction >= 0)
        elif isinstance(sort, list):
            sort_index = sort[0]
            reverse = None
    else:
        sort_index = None
        reverse = None

    if ids:
        skip = skip or 0
        end = skip + limit if limit else None
        ids = ids[skip:end]
        result = db[model][ids]
    else:
        result = db[model].query(criteria, skip=skip, limit=limit, sort_index=sort_index, reverse=reverse)

    return jsonpickle.encode(list(result), unpicklable=False)


@app.route("/<table_name>/_remove", methods=["GET", "POST"])
def remove(table_name):
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    criteria = req.get("criteria")
    ids = req.get("_id")
    if criteria:
        criteria = json.loads(req.get("criteria"))
        if isinstance(criteria, dict) and (len(criteria) == 1) and "_id" in criteria:
            ids = [c["$oid"] for c in criteria["_id"]]
        else:
            ids = None
            criteria = optimize(qj.compile(criteria))
        result = db[model].query(criteria)
    elif ids:
        ids = json.loads(ids)
        result = db[model][ids]
    else:
        return jsonify(ok=0)

    try:
        with transaction.manager:
            count = db.remove(result)
    except Exception, e:
        return jsonify(ok=0, message=str(e), error_type=e.__class__.__name__)

    return jsonify(ok=1, count=count)


@app.route("/<table_name>/_insert", methods=["GET", "POST"])
def insert(table_name):
    # POST has array of documents or one document
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "POST":
        try:
            if request.form.get("docs") is None:
                data = json.loads(request.data).get("docs")
            else:
                data = json.loads(request.form.get("docs"))
            if isinstance(data, dict):
                data = [data]
            objs = [model(**row) for row in data]
            with transaction.manager:
                oids = [{"$oid": db.add(o)} for o in objs]
            return jsonify(status={"ok": 1}, oids=oids)
        except Exception, e:
            return jsonify(ok=0, message=str(e), error_type=e.__class__.__name__)
    else:
        return jsonify(ok=0)


@app.route("/<table_name>/_get", methods=["GET", "POST"])
def get(table_name):
    db = dbs[session["username"]]
    model = getattr(models, table_name)

    if request.method == "GET":
        req = request.args
    elif request.method == "POST":
        req = request.form
    else:
        return jsonify(ok=0)

    data = json.loads(req["_id"])

    return jsonpickle.encode(db[model][data], unpicklable=False)


@app.route("/_disconnect", methods=["GET", "POST"])
def disconnect():
    if "username" in session:
        username = session.pop("username")
        if username in dbs:
            del dbs[username]
    return jsonify(ok=1)


def run(data_models=None, host=HOST, port=PORT, debug=DEBUG, secret_key=DEV_SECRET_KEY, zeo_socket=None, **kw):
    global models
    global zeosocket

    if isinstance(data_models, basestring):
        models = imp.load_source("models", data_models)
    else:
        models = data_models

    zeosocket = zeo_socket

    app.config["SECRET_KEY"] = secret_key
    app.run(host=host, port=port, debug=debug, **kw)


if __name__ == "__main__":
    run()



import itertools
import logging
from ZEO.ClientStorage import ClientStorage
from ZEO.Exceptions import ClientDisconnected


class ZEOBatchStorage(object):
    """
    ZEOStorage which loadEx object can return many oids at once
    """

    def loadBulk(self, oids):
        """
        Load multiple oids

        :param list oids: Iterable oids to load at once
        :return: Loaded oid objects
        :rtype: list
        """
        return [self.loadEx(oid) for oid in oids]


class BatchClientStorage(ClientStorage):
    """
    Allows to get objects in batches
    zlibstorage should wrap *after* this
    """

    def __init__(self, *args, **kw):
        super(BatchClientStorage, self).__init__(*args, **kw)
        self._load_oids = {}

    def _process_invalidations(self, tid, oids):
        for oid in oids:
            if self._load_oids.get(oid, None):
                self._load_oids[oid] = 0
            self._cache.invalidate(oid, tid)
        self._cache.setLastTid(tid)

        if self._db is not None:
            self._db.invalidate(tid, oids)

    def loadBulk(self, oids):
        """
        Storage API to return multiple objects
        We load a unique set of them, just in case

        :param list oids: Iterable oids to load at once
        :return: Loaded oid objects
        :rtype: list
        """
        # First, try to get whatever possible from cache
        self._load_lock.acquire()
        try:
            self._lock.acquire()    # for atomic processing of invalidations
            try:
                result = []
                for oid in oids:
                    out = self._cache.load(oid)
                    if not out:
                        self._load_oids[oid] = 1
                    else:
                        result.append(out)
            finally:
                self._lock.release()
            if len(self._load_oids) == 0:
                return result
            # If we ever get here, we need to load some more stuff
            # self._load_oids dictionary is protected by self._load_lock

            if self._server is None:
                raise ClientDisconnected()

            load_oids = self._load_oids.keys()

            # [(data, tid), (data, tid), ...]
            bulk_data = self._server.rpc.call("loadBulk", load_oids)

            data_size = 0
            for oid, (data, tid) in itertools.izip(load_oids, bulk_data):
                data_size += len(data)
                self._lock.acquire()    # for atomic processing of invalidations
                try:
                    if self._load_oids[oid]:  # Update cache only when there was no invalidation
                        self._cache.store(oid, tid, None, data)
                    del self._load_oids[oid]
                    result.append((data, tid))  # XXX shouldn't we provide a recent value from cache then?
                finally:
                    self._lock.release()
            logging.debug("Bulk-loaded {0} objects of size {1}".format(len(load_oids), data_size))
        finally:
            self._load_lock.release()

        return result



import BTrees
import hashlib
import persistent
import transaction
import ZODB

from os import path
from ZODB import FileStorage

from zerodb.crypto import rand
from zerodb.intid import IdStore as BaseIdStore


class IdStore(BaseIdStore):
    family = BTrees.family32


class User(persistent.Persistent):
    """
    Persistent class to store users
    """

    def __init__(self, username, pubkey, realm="ZEO", administrator=False, root=None):
        """
        :param str username: Username
        :param str pubkey: ECC public key
        :param str realm: Authentication realm
        :param bool administrator: Whether it is a super-user (can create
            trees for others)
        :param str root: Oid of root of the tree. Could be None or 8-byte string
        """
        self.username = username
        self.pubkey = pubkey
        self.realm = realm
        self.administrator = administrator
        self.root = root


def session_key(h_up, nonce):
    return hashlib.sha512(h_up + ":" + nonce).digest()


class PermissionsDatabase(object):
    """
    Class which represent permission storages:
    Config file with root users and ZODB db with others.

    ZODB db consists of two trees:
    str -> int: username -> user uid
    int -> obj: IdStore user uid -> User object

    When database is empty, we populate it with data for all users from
    config file.
    """
    realm = None
    family = BTrees.family32
    uid_pack = "i"  # family32 -> "i", family64 -> "q"

    def __init__(self, filename, realm=None):
        """
        :param str filename: Config file with users and their password hashes
        :param str realm: ZODB's default permissions realm
        """
        self.storage_filename = path.splitext(filename)[0] + ".db"
        self.storage = FileStorage.FileStorage(self.storage_filename)
        self.db = ZODB.DB(self.storage)
        self.db_conn = self.db.open()
        self.db_root = self.db_conn.root()
        root = self.db_root
        with transaction.manager:
            if not "users" in root:
                root["users"] = IdStore()  # uid -> user
            if not "usernames" in root:
                root["usernames"] = self.family.OI.BTree()  # username -> uid

        self.filename = filename
        self.load()

        # Frankly speaking, this realm-based security is questionable
        # Keep it here for now
        if realm:
            if self.realm and self.realm != realm:
                raise ValueError("Specified realm %r differs from database "
                                 "realm %r" % (realm or '', self.realm))
            else:
                self.realm = realm

        self.noncekey = rand(32)

    def close(self):
        """
        Close connection to the database
        """
        self.db_conn.close()
        self.storage.close()

    def load(self):
        """
        Load config file and fill the database if necessary.
        Public key is updated if changed in file.
        If user is removed from file, it is left in the database.
        All users from the file are administrators.
        """
        filename = self.filename
        if not filename:
            return

        if not path.exists(filename):
            return

        fd = open(filename, "r")

        # If first line is realm, set it. Othewise they are all users
        try:
            line = fd.next()
            if line.startswith("realm "):
                self.realm = line.strip().split(" ", 1)[1]
            else:
                fd.seek(0)
        except StopIteration:
            return

        users = self.db_root["users"]
        usernames = self.db_root["usernames"]

        with transaction.manager:
            for line in fd:
                username, pub = line.strip().split(":", 1)
                username = username.strip()
                pub = pub.strip().decode("hex")
                if username not in usernames:
                    user = User(username, pub, self.realm, administrator=True)
                    uid = users.add(user)
                    usernames[username] = uid
                else:
                    uid = usernames[username]
                    users[uid].pub = pub
                    users[uid].realm = self.realm

    def save(self, fd=None):
        """No need to save to the file. Compatibility with how ZODB does it"""
        pass

    def add_user(self, username, pubkey, administrator=False):
        """
        Add user to the database

        :param str username: Username
        :param str pubkey: ECC public key

        :raises LookupError: if username already exists
        """
        users = self.db_root["users"]
        usernames = self.db_root["usernames"]
        if username in usernames:
            raise LookupError("User %s already exists" % username)
        with transaction.manager:
            user = User(username, pubkey, self.realm,
                    administrator=administrator)
            uid = users.add(user)
            usernames[username] = uid

    def del_user(self, username):
        """
        Remove existing user from the db (only ZODB)

        :param str username: Username
        :raises LookupError: if user does not exist
        """
        users = self.db_root["users"]
        usernames = self.db_root["usernames"]
        if username not in usernames:
            raise LookupError("No such user: %s" % username)
        with transaction.manager:
            uid = usernames[username]
            del users[uid]
            del usernames[username]

    def change_key(self, username, pubkey):
        """
        Change password of existing user

        :param str username: Username
        :param str pubkey: ECC public key

        :raises LookupError: if user does not exist
        """
        users = self.db_root["users"]
        usernames = self.db_root["usernames"]
        if username not in usernames:
            raise LookupError("No such user: %s" % username)
        with transaction.manager:
            uid = usernames[username]
            users[uid].pubkey = pubkey

    def __getitem__(self, username):
        """
        :param str username: Username
        :return: User object
        :rtype: User
        :raises LookupError: if user does not exist
        """
        users = self.db_root["users"]
        usernames = self.db_root["usernames"]
        uid = usernames[username]
        if username not in usernames:
            raise LookupError("No such user: %s" % username)
        return users[uid]



"""
Module for auth with elliptic curve cryptography
"""

import hashlib
import struct
from ZEO import auth
from ZEO.auth.base import Client as BaseClient
from ZEO.auth import register_module
from ZEO.Exceptions import AuthError

import base
import subdb
from zerodb.crypto import rand
from zerodb.crypto import ecc


__module_name__ = "ecc_auth"


class ServerStorageMixin(object):

    def auth_get_challenge(self):
        """Return realm, challenge, and nonce."""
        self._challenge = rand(32)
        self._key_nonce = self._get_nonce()
        return self.auth_realm, self._challenge, self._key_nonce

    def auth_response(self, resp):
        # verify client response
        username, challenge, resp_sig = resp

        assert self._challenge == challenge

        user = self.database[username]
        verkey = ecc.public(user.pubkey)

        h_up = hashlib.sha256("%s:%s:%s" % (username, self.database.realm, user.pubkey)).digest()

        # regeneration resp from user, password, and nonce
        check = hashlib.sha256("%s:%s" % (h_up, challenge)).digest()
        verify = verkey.verify(resp_sig, check)
        if verify:
            self.connection.setSessionKey(base.session_key(h_up, self._key_nonce))
        # This class is per-connection, so we're safe to assign attributes
        authenticated = self._finish_auth(verify)
        if authenticated:
            user_id = self.database.db_root["usernames"][username]
            self.user_id = struct.pack(self.database.uid_pack, user_id)
        return authenticated


class StorageClass(ServerStorageMixin, subdb.StorageClass):
    pass


class Client(BaseClient):
    extensions = ["auth_get_challenge", "auth_response", "get_root_id"]

    def start(self, username, realm, password):
        priv = ecc.private(password)
        _realm, challenge, nonce = self.stub.auth_get_challenge()
        # _realm is str, challenge is 32-byte hash, nonce as well
        if _realm != realm:
            raise AuthError("expected realm %r, got realm %r"
                            % (_realm, realm))
        h_up = hashlib.sha256("%s:%s:%s" % (username, realm, priv.get_pubkey())).digest()

        check = hashlib.sha256("%s:%s" % (h_up, challenge)).digest()
        sig = priv.sign(check)
        result = self.stub.auth_response((username, challenge, sig))
        if result:
            return base.session_key(h_up, nonce)
        else:
            return None


def register_auth():
    if __module_name__ not in auth._auth_modules:
        register_module(__module_name__, StorageClass, Client, base.PermissionsDatabase)



from StringIO import StringIO
from zerodb.catalog import query_json as qj
from zerodb.catalog.query import optimize
from zerodb.catalog.query import Gt, Contains
from zerodb.catalog.query import Not, Eq, InRange


def qtree(q):
    q = optimize(q)
    f = StringIO()
    q.print_tree(out=f)
    f.seek(0)
    out = f.read()
    f.close()
    return out


def test_query_json():
    q1 = qj.compile({"$and": [{"field1": {"$gt": 10}}, {"field2": {"$eq": "hello"}}]})
    q2 = Gt("field1", 10) & Eq("field2", "hello")
    assert qtree(q1) == qtree(q2)

    q1 = qj.compile({"$not": {"$or": [{"field1": {"$range": [1, 15]}}, {"field2": {"$text": "hello"}}]}})
    q2 = Not(InRange("field1", 1, 15) | Contains("field2", "hello"))
    assert qtree(q1) == qtree(q2)



import ctypes
import multiprocessing
import threading

from time import sleep

from db import Page

id_conn_child = None


def test_thread_pooling(db):
    id_conn_parent = id(db._connection)

    def f():
        global id_conn_child

        assert len(db[Page]) > 0  # connect automatically
        id_conn_child = id(db._connection)
        sleep(0.2)

    thread = threading.Thread(target=f)
    thread.start()
    sleep(0.1)
    assert len(db._db.pool.all) - len(db._db.pool.available) == 2

    thread.join()

    # Threadlocals die only when another thread accesses them
    db._root

    # TODO: close connection when thread dies
    assert len(db._db.pool.all) - len(db._db.pool.available) == 1

    # Test that child used connection other than parent
    assert id_conn_child != id_conn_parent
    assert id_conn_child is not None


def test_forking(db):
    id_conn_parent = id(db._connection)
    id_conn_child = multiprocessing.Value(ctypes.c_int64)

    def f():
        assert len(db[Page]) > 0  # Check that DB is functional in the subprocess
        id_conn_child.value = id(db._connection)

    p = multiprocessing.Process(target=f)
    p.start()
    p.join()

    # Test that child used connection other than parent
    assert id_conn_child.value != 0
    assert id_conn_child.value != id_conn_parent



from zerodb.catalog.indexes.text import CatalogTextIndex
from zerodb.catalog.indexes.field import CatalogFieldIndex
import exceptions


class Indexable(object):
    Index = None

    def __init__(self, default=None, virtual=None, index=True):
        """
        :param default: Default value (which can be callable, like utcnow)
        :param virtual: Virtual value which is *only* calculated but is not
            stored. Still, it is indexed
        :param bool index: If False, we just use schema for validation
        """
        self.default = default
        self.virtual = virtual
        self.indexed = index

        if (default is not None) and (virtual is not None):
            raise exceptions.FieldException("One cannot simultaneously set the default value"
                    "and claim that the field is derived by calculation only")

    def __repr__(self):
        return "Indexable field <%s>" % self.__class__.__name__


class Field(Indexable):
    """
    Field of any type which supports comparisions
    """
    Index = CatalogFieldIndex


class Text(Indexable):
    """
    Text field to be used for fulltext search
    """
    Index = CatalogTextIndex



from repoze.catalog import query
from zerodb import trees
from zerodb.util.iter import Sliceable, ListPrefetch


def _to_set(flavor, data):
    if isinstance(data, Sliceable):
        return flavor.Set(data)
    else:
        return data


class LogicMixin:
    def __and__(self, right):
        self._check_type("and", right)
        return And(self, right)

    def __or__(self, right):
        self._check_type("or", right)
        return Or(self, right)


class Query(LogicMixin, query.Query):
    pass


class Comparator(LogicMixin, query.Comparator):
    pass


class Contains(LogicMixin, query.Comparator):
    """Contains query.

    CQE equivalent: 'foo' in index
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyContains(self._get_value(names))

    def __str__(self):
        return '%s in %s' % (repr(self._value), self.index_name)

    def negate(self):
        return DoesNotContain(self.index_name, self._value)


class DoesNotContain(LogicMixin, query.Comparator):
    """CQE equivalent: 'foo' not in index
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyDoesNotContain(self._get_value(names))

    def __str__(self):
        return '%s not in %s' % (repr(self._value), self.index_name)

    def negate(self):
        return Contains(self.index_name, self._value)


class Eq(LogicMixin, query.Eq):
    pass


class NotEq(LogicMixin, query.NotEq):
    pass


class Gt(Comparator):
    """ Greater than query.

    CQE equivalent: index > 'foo'
    """
    operator = '>'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyGt(self._get_value(names))

    def negate(self):
        return Le(self.index_name, self._value)


class Lt(Comparator):
    """ Less than query.

    CQE equivalent: index < 'foo'
    """
    operator = '<'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyLt(self._get_value(names))

    def negate(self):
        return Ge(self.index_name, self._value)


class Ge(Comparator):
    """Greater (or equal) query.

    CQE equivalent: index >= 'foo'
    """
    operator = '>='

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyGe(self._get_value(names))

    def negate(self):
        return Lt(self.index_name, self._value)


class Le(Comparator):
    """Less (or equal) query.

    CQE equivalent: index <= 'foo
    """
    operator = '<='

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyLe(self._get_value(names))

    def negate(self):
        return Gt(self.index_name, self._value)


class InRange(LogicMixin, query.InRange):
    """ Index value falls within a range.

    CQE eqivalent: lower < index < upper
                   lower <= index <= upper
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyInRange(
            self._get_start(names), self._get_end(names),
            self.start_exclusive, self.end_exclusive)

    def negate(self):
        return NotInRange(self.index_name, self._start, self._end,
                          self.start_exclusive, self.end_exclusive)


class NotInRange(LogicMixin, query.NotInRange):
    """ Index value falls outside a range.

    CQE eqivalent: not(lower < index < upper)
                   not(lower <= index <= upper)
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyNotInRange(
            self._get_start(names), self._get_end(names),
            self.start_exclusive, self.end_exclusive)

    def negate(self):
        return InRange(self.index_name, self._start, self._end,
                       self.start_exclusive, self.end_exclusive)


class Any(LogicMixin, query.Any):
    pass


class NotAny(LogicMixin, query.NotAny):
    pass


class All(LogicMixin, query.All):
    pass


class NotAll(LogicMixin, query.NotAll):
    pass


class BoolOp(LogicMixin, query.BoolOp):
    family = trees.family32


class Or(LogicMixin, query.Or):
    family = trees.family32

    def _apply(self, catalog, names):
        # XXX Or query can be done lazily using heapq's sorted join
        IF = self.family.IF
        queries = self.queries
        result = queries[0]._apply(catalog, names)
        for q in queries[1:]:
            next_result = q._apply(catalog, names)
            if len(result) == 0:
                result = next_result
            elif len(next_result) > 0:
                _, result = self.family.IF.weightedUnion(_to_set(IF, result), _to_set(IF, next_result))
        return result


class And(LogicMixin, query.And):
    family = trees.family32

    def _apply(self, catalog, names):
        # XXX figure out doing smallest part first for performance
        IF = self.family.IF
        queries = self.queries
        result = queries[0]._apply(catalog, names)
        for q in queries[1:]:
            if len(result) == 0:
                return IF.Set()
            next_result = q._apply(catalog, names)
            if len(next_result) == 0:
                return IF.Set()
            _, result = IF.weightedIntersection(_to_set(IF, result), _to_set(IF, next_result))
        return result


class Not(LogicMixin, query.Not):
    pass


class Name(LogicMixin, query.Name):
    pass

optimize = query.optimize
parse_query = query.parse_query



import pickle
from StringIO import StringIO


class DebugUnpickler(pickle.Unpickler):
    """
    Unpickler which returns class names instead of unpickling (for debug purposes)
    """

    def find_class(self, module, name):
        return name

DebugUnpickler.dispatch[pickle.REDUCE] = lambda x: None


def debug_loads(obj):
    up = DebugUnpickler(StringIO(obj))
    return up.load()



import persistent
from zope.index.text import widcode
from zerodb.catalog.indexes import pwid


def test_wid():
    test_wids = range(5)
    pw = pwid.PersistentWid.encode_wid(test_wids)
    assert isinstance(pw, persistent.Persistent)
    assert pw.decode_wid() == test_wids
    assert pw.find(widcode.encode([1]))



from setuptools import setup, find_packages

INSTALL_REQUIRES = [
    'BTrees',
    'zope.component',
    'zodbpickle',
    'ZODB',
    'zope.index',
    'repoze.catalog',
    'cachetools',
    'zc.zlibstorage',
    'pycryptodome',
    'flask-cors',
    'flask',
    'requests',
    'jsonpickle',
    'pyelliptic',
    'ecdsa',
    'zope.event',
    'zope.lifecycleevent',
    'aes256gcm-nacl']

setup(
    name="zerodb",
    version="0.96.21",
    description="End-to-end encrypted database",
    author="ZeroDB Inc.",
    author_email="michael@zerodb.io",
    license="AGPLv3",
    url="http://zerodb.io",
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
)



import pytest
from pickle import loads
from zerodb.storage import client_storage
from db import TEST_PASSPHRASE


@pytest.fixture(scope="module")
def zeo_storage(request, zeo_server):
    return client_storage(zeo_server,
            username="root", password=TEST_PASSPHRASE, realm="ZERO", debug=True)


def test_loadBulk(zeo_storage):
    # oids in ZODB start from zero, and there are less than 10
    root_id, _ = zeo_storage.get_root_id()
    count_0 = zeo_storage._debug_download_count
    oids = [root_id]
    out = zeo_storage.loadBulk(oids)
    count_1 = zeo_storage._debug_download_count
    assert count_1 - count_0 == 1

    assert len(out) == len(oids)

    for dump, oid in out:
        assert loads(dump) is not None

    for oid in oids:
        assert zeo_storage._cache.load(oid)

    for oit in oids:
        zeo_storage.load(oid)
    assert zeo_storage._debug_download_count == count_1



from repoze.catalog.indexes.common import *

_marker = ()


class CallableDiscriminatorMixin(object):
    """
    Compatibility function which makes index pickleable
    """
    def _init_discriminator(self, discriminator):
        if isinstance(discriminator, tuple):
            self.discriminator, = discriminator
            self.discriminator_callable = True
        else:
            if not isinstance(discriminator, basestring):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
            self.discriminator = discriminator
            self.discriminator_callable = False

    def index_doc(self, docid, obj):
        if self.discriminator_callable:
            # Model class definition has a list of virtual fields
            virtuals = getattr(obj.__class__, "_z_virtual_fields", {})
            value = virtuals.get(self.discriminator, _marker)
            if value != _marker:
                try:
                    value = value(obj)
                except:
                    value = _marker
        else:
            value = getattr(obj, self.discriminator, _marker)

        if value is _marker:
            # unindex the previous value
            super(CatalogIndex, self).unindex_doc(docid)

            # Store docid in set of unindexed docids
            self._not_indexed.add(docid)

            return None

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object %s' %
                             value)

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object %s' %
                             value)

        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)

        return super(CatalogIndex, self).index_doc(docid, value)



import persistent
import fields
import exceptions
from zope.lifecycleevent import modified
from zerodb.intid import IdStore
from zerodb.trees import family32
from zerodb.catalog import Catalog
from zerodb.models.fields import Field, Text


class ModelMeta(type):
    def __init__(cls, name, bases, dct):
        if bases != (persistent.Persistent,):  # Only subclasses of Model can do it, not Model itself
            used_fields = set(filter(lambda key:
                    not key.startswith("_") and
                    isinstance(dct[key], fields.Indexable),
                dct.keys()))

            required_fields = set(filter(lambda key: dct[key].default is None, used_fields))
            virtual_fields = {key: dct[key].virtual for key in used_fields
                    if dct[key].virtual is not None}

            default_fields = used_fields - required_fields

            cls._z_indexed_fields = set(filter(lambda key: dct[key].indexed, used_fields))
            cls._z_default_fields = default_fields.difference(virtual_fields)
            cls._z_virtual_fields = virtual_fields
            cls._z_required_fields = required_fields.difference(virtual_fields)

            cls.__modelname__ = cls.__modelname__ or name.lower()

            # Clean up all the fields we've used
            dct = {k: dct[k] for k in dct.keys() if k not in used_fields}

        super(ModelMeta, cls).__init__(name, bases, dct)


class Model(persistent.Persistent):
    """
    Data model to easily create indexable persistent objects.
    If an object declares a property from indexable, this property is indexed.
    *All* other properties are stored but unindexed

    Example:
        >>> class Page(Model):
        ...     title = indexable.Field()
        ...     text = indexable.Text()

        ... page = Page(title="Hello", text="World", extra=12345)
    """

    # TODO we should (?) provide some way to save model definition into the database
    # and generate model class from there.
    # This could be valuable when working with json which doesn't keep any python code
    # Or we should make a method which allows to generate a new model from json

    __metaclass__ = ModelMeta
    __modelname__ = None
    __family__ = family32

    def __init__(self, **kw):
        missed_fields = self._z_required_fields.difference(kw)
        if missed_fields:
            raise exceptions.ModelException("You should provide fields: " + ", ".join(map(str, missed_fields)))

        for field in self._z_default_fields.difference(kw):
            default = getattr(self.__class__, field).default
            if callable(default):
                default = default()
            setattr(self, field, default)

        for field, value in kw.iteritems():
            setattr(self, field, value)

    def __setattr__(self, name, value):
        origattr = getattr(self, name, None)
        if origattr is not None and name in self._z_indexed_fields and \
                not isinstance(origattr, fields.Indexable) and hasattr(self, "_p_uid"):  # reindex notify
            modified(self, name)
        super(Model, self).__setattr__(name, value)

    @classmethod
    def create_store(cls):
        """Returns intid.IdStore storage for this class"""
        return IdStore(family=cls.__family__)

    @classmethod
    def create_catalog(cls):
        """Creates and returns catalog for this model with indexes in it"""
        catalog = Catalog()  # TODO: specify family as a parameter
        for name in cls._z_indexed_fields:
            field = getattr(cls, name)

            if field.virtual:
                # Tuple discriminator means virtual field.
                # Index has to look it up in model._z_virtual_fields,
                # where model = obj.__class__
                discriminator = (name,)
            else:
                discriminator = None

            catalog[name] = field.Index(discriminator or name)
        return catalog



from zerodb.transform import init_crypto, encrypt, decrypt
from zerodb.transform.encrypt_aes import AES256Encrypter


def test_utilities():
    AES256Encrypter.register_class(default=True)

    init_crypto(passphrase="Hello world")

    test_text = "Test text " * 100

    encrypted_text = encrypt(test_text)
    assert encrypted_text.startswith(".e")
    assert decrypt(encrypted_text) == test_text



import itertools
from BTrees.Length import Length
from zope.index.text import TextIndex as ZopeTextIndex
from zope.index.text.lexicon import CaseNormalizer
from zope.index.text.lexicon import Lexicon as _Lexicon
from zope.index.text.lexicon import Splitter
from zope.index.text.lexicon import StopWordRemover
from zope.index.text.okapiindex import OkapiIndex as _OkapiIndex
from repoze.catalog.indexes.text import CatalogTextIndex as _CatalogTextIndex
from zerodb import trees
from zerodb.catalog.indexes.common import CallableDiscriminatorMixin
from zerodb.storage import prefetch_trees, parallel_traversal
from zerodb.catalog.indexes.pwid import PersistentWid


def _text2list(text):
    # Helper: splitter input may be a string or a list of strings
    # Fixed from zope.index.text.lexicon
    try:
        text + u""
    except UnicodeDecodeError:
        try:
            return [text.decode("utf-8")]
        except UnicodeDecodeError:
            # It wasn't utf-8. But at least, it was text!
            return [text]
    except TypeError:
        return text
    else:
        return [text]


class Lexicon(_Lexicon):
    family = trees.family32  # In comparison with standard Lexicon, use bigger buckets

    def __init__(self, *pipeline):
        self._wids = self.family.OI.BTree()
        self._words = self.family.IO.BTree()
        self.wordCount = Length()
        self._pipeline = pipeline

    def sourceToWordIds(self, text):
        if text is None:
            text = ''
        last = _text2list(text)
        for element in self._pipeline:
            last = element.process(last)
        if not isinstance(self.wordCount, Length):
            # Make sure wordCount is overridden with a BTrees.Length.Length
            self.wordCount = Length(self.wordCount())
        # Strategically unload the length value so that we get the most
        # recent value written to the database to minimize conflicting wids
        # Because length is independent, this will load the most
        # recent value stored, regardless of whether MVCC is enabled
        self.wordCount._p_deactivate()
        parallel_traversal(self._wids, last)
        return list(map(self._getWordIdCreate, last))


class OkapiIndex(_OkapiIndex):

    def clear(self):
        # wid -> {docid -> weight}; t -> D -> w(D, t)
        self._wordinfo = trees.family32.IO.BTree()
        # XXX
        # Scalability of this Zope's approach is pretty bad (esp. when many documents).
        # Following is needed:
        # _wordinfo = BTree of (word, weight, docid) - TreeSet could be used instead
        # searching a keyword will be as _wordinfo.keys((word_start, None, None)),
        # already sorted by weight (just has to be multiplied by idf)
        # this works for both search and glob_search
        # However, when searching multiple keywords,
        # Need to find an efficient (logarithmic) algorithm of
        # incremental, weighted set intersection
        # Even without efficient intersection, it is faster and more secure anyway
        # XXX

        # docid -> weight
        self._docweight = self.family.IF.BTree()

        # docid -> WidCode'd list of wids (~1/4 of document size)
        # Used for un-indexing, and for phrase search.
        self._docwords = self.family.IO.BTree()

        # Use a BTree length for efficient length computation w/o conflicts
        self.wordCount = Length()
        self.documentCount = Length()

    def _mass_add_wordinfo(self, wid2weight, docid):
        dicttype = type({})
        # self._wordinfo - IOBTree of docid -> weight trees
        get_doc2score = self._wordinfo.get
        new_word_count = 0

        # Fill up cache for performance over the network
        wids = wid2weight.keys()
        parallel_traversal(self._wordinfo, wids)
        parallel_traversal(map(get_doc2score, wids), [docid] * len(wids))

        for wid, weight in wid2weight.items():
            doc2score = get_doc2score(wid)
            if doc2score is None:
                doc2score = {}
                new_word_count += 1
            elif (isinstance(doc2score, dicttype) and
                  len(doc2score) == self.DICT_CUTOFF):
                doc2score = self.family.IF.BTree(doc2score)
            doc2score[docid] = weight
            self._wordinfo[wid] = doc2score  # not redundant:  Persistency!
        try:
            self.wordCount.change(new_word_count)
        except AttributeError:
            # upgrade wordCount to Length object
            self.wordCount = Length(len(self._wordinfo))

    def index_doc(self, docid, text):
        if docid in self._docwords:
            return self._reindex_doc(docid, text)
        wids = self._lexicon.sourceToWordIds(text)
        wid2weight, docweight = self._get_frequencies(wids)
        self._mass_add_wordinfo(wid2weight, docid)
        self._docweight[docid] = docweight
        self._docwords[docid] = PersistentWid.encode_wid(wids)
        try:
            self.documentCount.change(1)
        except AttributeError:
            # upgrade documentCount to Length object
            self.documentCount = Length.Length(len(self._docweight))
        count = len(wids)
        self._change_doc_len(count)
        return count

    def _reindex_doc(self, docid, text):
        # Touch as few docid->w(docid, score) maps in ._wordinfo as possible.
        self._change_doc_len(-self._docweight[docid])

        old_wids = self.get_words(docid)
        old_wid2w, old_docw = self._get_frequencies(old_wids)

        new_wids = self._lexicon.sourceToWordIds(text)
        new_wid2w, new_docw = self._get_frequencies(new_wids)

        old_widset = self.family.IF.TreeSet(old_wid2w.keys())
        new_widset = self.family.IF.TreeSet(new_wid2w.keys())

        IF = self.family.IF
        in_both_widset = IF.intersection(old_widset, new_widset)
        only_old_widset = IF.difference(old_widset, in_both_widset)
        only_new_widset = IF.difference(new_widset, in_both_widset)
        del old_widset, new_widset

        for wid in only_old_widset.keys():
            self._del_wordinfo(wid, docid)

        for wid in only_new_widset.keys():
            self._add_wordinfo(wid, new_wid2w[wid], docid)

        for wid in in_both_widset.keys():
            # For the Okapi indexer, the "if" will trigger only for words
            # whose counts have changed.  For the cosine indexer, the "if"
            # may trigger for every wid, since W(d) probably changed and
            # W(d) is divided into every score.
            newscore = new_wid2w[wid]
            if old_wid2w[wid] != newscore:
                self._add_wordinfo(wid, newscore, docid)

        self._docweight[docid] = new_docw
        self._docwords[docid] = PersistentWid.encode_wid(new_wids)
        return len(new_wids)

    def get_words(self, docid):
        """Return a list of the wordids for a given docid."""
        return self._docwords[docid].decode_wid()

    def _search_wids(self, wids):
        # Bulk-fetch all the info we want to use
        if len(wids) > 1:
            parallel_traversal(self._wordinfo, wids)
        prefetch_trees([self._wordinfo[wid] for wid in wids])

        docids = list(set(itertools.chain(
            *[self._wordinfo[wid].keys() for wid in wids])))
        if len(docids) > 1:
            parallel_traversal(self._docweight, docids)

        return super(OkapiIndex, self)._search_wids(wids)


class CatalogTextIndex(CallableDiscriminatorMixin, _CatalogTextIndex):
    family = trees.family32

    def __init__(self, discriminator, lexicon=None, index=None):
        self._init_discriminator(discriminator)

        self._not_indexed = self.family.IF.Set()

        lexicon = lexicon or Lexicon(Splitter(), CaseNormalizer(), StopWordRemover())
        index = index or OkapiIndex(lexicon, family=self.family)

        ZopeTextIndex.__init__(self, lexicon, index)
        self.clear()



import hashlib
import pyelliptic as pe  # This module re-uses OPENSSL's code
import ecdsa  # But this one is more functional :-)

# We can get rid of this pure-python ecdsa module and use compressed keys
# For OpenSSL-based implementation see:
#   https://github.com/Bitmessage/PyBitmessage/blob/master/src/highlevelcrypto.py
# For pure python implementation (just in case!) see:
#   https://github.com/vbuterin/pybitcointools/


# We use curve standard for Bitcoin by default
CURVE = "secp256k1"
CURVE_ecdsa = ecdsa.SECP256k1


def priv2pub(priv):
    """
    Converts 32-bytes private key to public key

    :param str priv: 32-byte private key
    :return: (pub_x, pub_y)
    :rtype: tuple
    """
    pub = ecdsa.SigningKey.from_string(priv, curve=CURVE_ecdsa).get_verifying_key().to_string()
    return pub[:32], pub[32:]


def private(passphrase):
    priv = hashlib.sha256("elliptic" + hashlib.sha256(passphrase).digest()).digest()
    px, py = priv2pub(priv)
    key = pe.ECC(curve=CURVE)
    key._set_keys(px, py, priv)
    return key


def public(pub):
    return pe.ECC(curve=CURVE, pubkey=pub)



import pytest
from datetime import datetime
from datetime import timedelta

from zerodb import models
import zerodb
import zerodb.db
from zerodb.models import fields
from zerodb.models.exceptions import ModelException
from db import TEST_PASSPHRASE


class TestMe(models.Model):
    title = fields.Field(index=False)
    content = fields.Text(index=False)
    age = fields.Field(default=0)
    timestamp = fields.Field(default=datetime.utcnow)
    all_text = fields.Text(virtual=lambda o: o.title + " " + o.content)


def test_model_metaclass():
    assert len(TestMe._z_indexed_fields) == 3
    assert len(TestMe._z_required_fields) == 2
    assert len(TestMe._z_default_fields) == 2
    assert TestMe._z_virtual_fields.keys() == ["all_text"]


def test_model():
    test_timestamp = datetime(year=1984, month=1, day=1)
    obj1 = TestMe(title="Hello", content="World", age=5, timestamp=test_timestamp)
    obj2 = TestMe(title="Hello world", content="Lorem ipsum")
    obj3 = TestMe(title="Hello", content="World", extra="something")

    with pytest.raises(ModelException):
        TestMe(title="This is not enough")

    assert obj1.title == "Hello"
    assert obj1.content == "World"
    assert obj1.age == 5
    assert obj1.timestamp == test_timestamp

    assert obj1._z_virtual_fields["all_text"](obj1) == "Hello World"

    assert obj2.title == "Hello world"
    assert obj2.content == "Lorem ipsum"
    assert obj2.age == 0
    assert (datetime.utcnow() - obj2.timestamp) < timedelta(seconds=1)

    assert obj1.title == "Hello"
    assert obj1.content == "World"
    assert obj3.age == 0
    assert (datetime.utcnow() - obj3.timestamp) < timedelta(seconds=1)
    assert obj3.extra == "something"


def test_db(zeo_server):
    db = zerodb.DB(zeo_server, username="root", password=TEST_PASSPHRASE, debug=True)
    assert len(db._models) == 0
    assert isinstance(db[TestMe], zerodb.db.DbModel)
    assert len(db._models) == 1
    assert TestMe in db._models
    db.disconnect()


def test_dbmodel(zeo_server):
    db = zerodb.DB(zeo_server, username="root", password=TEST_PASSPHRASE, debug=True)
    assert db[TestMe]._model == TestMe
    assert db[TestMe]._db == db
    assert db[TestMe]._catalog_name == "catalog__testme"
    assert db[TestMe]._intid_name == "store__testme"
    db.disconnect()



from query import Eq, Lt, Gt, And, Or, Not, InRange, Contains,\
                  NotEq, Le, Ge, NotInRange, DoesNotContain,\
                  Any, All, NotAny, NotAll

"""
Query examples (similar to http://docs.mongodb.org/manual/reference/operator/query/):

    {"$and": [{"field1": {"$gt": 10}}, {"field2": {"$text": "hello"}}]}
    {field: {"$range": [1, 10]}}
"""

logical_operators = {
        "$and": And,
        "$or": Or,
        "$not": Not}

field_operators = {
        "$eq": Eq,
        "$ne": NotEq,
        "$lt": Lt,
        "$lte": Le,
        "$gt": Gt,
        "$gte": Ge,
        "$range": InRange,
        "$nrange": NotInRange,
        "$text": Contains,
        "$ntext": DoesNotContain,
        "$in": Any,
        "$all": All,
        "$nany": NotAny,
        "$nin": NotAll}


def compile(q):
    """
    :param dict q: deserialized json
    :returns: query object
    :rtype: zerodb.catalog.query.Query
    """
    assert len(q) == 1
    key = q.keys()[0]

    if key in logical_operators:
        if isinstance(q[key], list):
            return logical_operators[key](*map(compile, q[key]))
        else:
            return logical_operators[key](compile(q[key]))
    else:
        assert isinstance(q[key], dict)
        assert len(q[key]) == 1
        opkey = q[key].keys()[0]
        params = q[key][opkey]
        if not isinstance(params, list):
            params = [params]
        return field_operators[opkey](key, *params)



from cStringIO import StringIO
from hashlib import sha256

from encrypt_common import CommonEncrypter
from Crypto.Random import get_random_bytes as rand
from zerodb.crypto.exceptions import WrongKeyError

import logging

try:
    from zerodbext.aead import AES256GCM
    from zerodbext.aead.bindings import crypto_aead_aes256gcm_ABYTES as MAC_LEN
    import zerodbext.aead.exceptions
    CryptoError = (zerodbext.aead.exceptions.CryptoError, ValueError)
    use_sodium = True

except ImportError:
    CryptoError = ValueError
    use_sodium = False

finally:
    from Crypto.Cipher import AES
    if hasattr(AES, "MODE_GCM"):
        MODE = AES.MODE_GCM
    else:
        if use_sodium:
            logging.info("Old AES256 with IV_SIZE=16 will not work")
        else:
            raise ImportError("You need to have PyCryptoDome which has support for AES256-GCM")


class AES256Encrypter(CommonEncrypter):
    """
    AES256 with key derived as sha256(passphrase)
    """
    name = "AES256-GCM"
    attributes = ("passphrase", "key")
    iv_size = 12
    key_size = 32

    use_sodium = use_sodium

    def _init_encryption(self, passphrase=None, key=None):
        """
        :param str passphrase: Passphrase for encryption (use if no key is set)
        :param str key: Symmetric key (if passphrase is not set)
        """
        if passphrase is not None:
            assert key is None
            key = sha256(passphrase).digest()
        elif key is not None:
            assert len(key) == self.key_size

        if self.use_sodium:
            self._box = AES256GCM(key)
        else:
            self.key = key

    def _encrypt(self, data):
        """
        :param str data: Data to encrypt
        :return: Encrypted data with hash inside and IV outside
        :rtype: str
        """
        iv = rand(self.iv_size)
        if self.use_sodium:
            edata, tag = self._box.encrypt_and_mac(data, iv)
        else:
            cipher = AES.new(self.key, MODE, iv)
            edata, tag = cipher.encrypt_and_digest(data)
        return iv + tag + edata

    def _decrypt(self, edata):
        """
        :param str edata: Data to decrypt
        :return: Decrypted data
        :rtype: str
        """
        f = StringIO(edata)
        iv = f.read(self.iv_size)
        if self.use_sodium:
            tag = f.read(MAC_LEN)
        else:
            cipher = AES.new(self.key, MODE, iv)
            tag = f.read(cipher._mac_len)
        data = f.read(-1)
        try:
            if self.use_sodium:
                return str(self._box.decrypt_and_verify(data, tag, iv))
            else:
                return cipher.decrypt_and_verify(data, tag)
        except CryptoError:
            raise WrongKeyError("Data couldn't be decrypted. Probably your key is wrong")


class AES256EncrypterV0(AES256Encrypter):
    name = "AES256"
    use_sodium = False
    iv_size = 16
